"""Initialize the app"""
from django.utils.translation import gettext_lazy as _

__version__ = "1.0.4"
__title__ = "Campaign"
__title_translated__ = _("AA Campaigns")

__esi_compatibility_date__ = "2025-11-06"

__package_name__ = "aa-campaign"
__app_name__ = "aa-campaign"
__app_name_verbose__ = "Alliance Auth Campaigns"
__app_name_useragent__ = "AA-Campaign"

__github_url__ = f"https://github.com/BroodLK/aa-campaigns"
